# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from typer import Typer

from .auth import app as auth_app
from .env import app as env_app
from .misc import cli_options
from .predict import predict
from .predictors import archive_predictor, create_predictor, delete_predictor, list_predictors, retrieve_predictor, search_predictors
from ..version import __version__

# Define CLI
app = Typer(
    name=f"Function CLI {__version__}",
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_short=True,
    add_completion=False
)

# Add top level options
app.callback()(cli_options)

# Add subcommands
app.add_typer(auth_app, name="auth", help="Login, logout, and check your authentication status.")
app.add_typer(env_app, name="env", help="Manage predictor environment variables.")

# Add top-level commands
app.command(name="create", help="Create a predictor.")(create_predictor)
app.command(name="delete", help="Delete a predictor.")(delete_predictor)
app.command(name="predict", help="Make a prediction.", context_settings={ "allow_extra_args": True, "ignore_unknown_options": True })(predict)
app.command(name="list", help="List predictors.")(list_predictors)
app.command(name="search", help="Search predictors.")(search_predictors)
app.command(name="retrieve", help="Retrieve a predictor.")(retrieve_predictor)
app.command(name="archive", help="Archive a predictor.")(archive_predictor)

# Run
if __name__ == "__main__":
    app()


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import asdict
from rich import print_json
from pathlib import Path
from typer import Argument, Option
from typing import List

from ..api import Acceleration, AccessMode, Predictor, PredictorStatus, PredictorType
from .auth import get_access_key

def retrieve_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    predictor = Predictor.retrieve(
        tag=tag,
        access_key=get_access_key()
    )
    predictor = asdict(predictor) if predictor else None
    print_json(data=predictor)

def list_predictors (
    owner: str=Option(None, help="Predictor owner. This defaults to the current user."),
    status: PredictorStatus=Option(PredictorStatus.Active, help="Predictor status. This defaults to `ACTIVE`."),
    offset: int=Option(None, help="Pagination offset."),
    count: int=Option(None, help="Pagination count.")
):
    predictors = Predictor.list(
        owner=owner,
        status=status,
        offset=offset,
        count=count,
        access_key=get_access_key()
    )
    predictors = [asdict(predictor) for predictor in predictors] if predictors is not None else None
    print_json(data=predictors)

def search_predictors (
    query: str=Argument(..., help="Search query."),
    offset: int=Option(None, help="Pagination offset."),
    count: int=Option(None, help="Pagination count.")
):
    predictors = Predictor.search(
        query=query,
        offset=offset,
        count=count,
        access_key=get_access_key()
    )
    predictors = [asdict(predictor) for predictor in predictors]
    print_json(data=predictors)

def create_predictor (
    tag: str=Argument(..., help="Predictor tag."),
    notebook: Path=Argument(..., help="Path to predictor notebook."),
    type: PredictorType=Option(None, case_sensitive=False, help="Predictor type. This defaults to `CLOUD`."),
    access: AccessMode=Option(None, case_sensitive=False, help="Predictor access mode. This defaults to `PRIVATE`."),
    description: str=Option(None, help="Predictor description. This must be less than 200 characters long."),
    media: Path=Option(None, help="Predictor image path."),
    acceleration: Acceleration=Option(None, case_sensitive=False, help="Cloud predictor acceleration. This defaults to `CPU`."),
    license: str=Option(None, help="Predictor license URL."),
    env: List[str]=Option([], help="Specify a predictor environment variable."),
    overwrite: bool=Option(None, "--overwrite", help="Overwrite any existing predictor with the same tag.")
):
    environment = { e.split("=")[0].strip(): e.split("=")[1].strip() for e in env }
    predictor = Predictor.create(
        tag=tag,
        notebook=notebook,
        type=type,
        access=access,
        description=description,
        media=media,
        acceleration=acceleration,
        environment=environment,
        license=license,
        overwrite=overwrite,
        access_key=get_access_key()
    )
    predictor = asdict(predictor)
    print_json(data=predictor)

def delete_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    result = Predictor.delete(
        tag=tag,
        access_key=get_access_key()
    )
    print_json(data=result)

def archive_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    predictor = Predictor.archive(
        tag=tag,
        access_key=get_access_key()
    )
    print_json(data=asdict(predictor))


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import asdict
from rich import print_json
from typer import Argument, Option, Typer

from ..api import EnvironmentVariable
from .auth import get_access_key

app = Typer(no_args_is_help=True)

@app.command(name="list", help="List environment variables.")
def list_envs (
    organization: str=Option(None, help="Organization username. Use this for organization environment variables."),
):
    environments = EnvironmentVariable.list(
        organization=organization,
        access_key=get_access_key()
    )
    environments = [asdict(env) for env in environments]
    print_json(data=environments)

@app.command(name="create", help="Create an environment variable.")
def create_env (
    name: str=Argument(..., help="Variable name."),
    value: str=Argument(..., help="Variable value."),
    organization: str=Option(None, help="Organization username. Use this for organization environment variables."),
):
    environment = EnvironmentVariable.create(
        name=name,
        value=value,
        organization=organization,
        access_key=get_access_key()
    )
    environment = asdict(environment)
    print_json(data=environment)

@app.command(name="delete", help="Delete an environment variable.")
def delete_env (
    name: str=Argument(..., help="Variable name."),
    organization: str=Option(None, help="Organization username. Use this for organization environment variables."),
):
    result = EnvironmentVariable.delete(
        name=name,
        organization=organization,
        access_key=get_access_key()
    )
    print_json(data=result)


#
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from dataclasses import asdict, dataclass
from numpy import ndarray
from pathlib import Path
from PIL import Image
from platform import system
from typing import Any, Dict, List, Union
from uuid import uuid4

from .api import query
from .predictor import PredictorType
from .value import Value

@dataclass(frozen=True)
class Prediction:
    """
    Prediction.

    Members:
        id (str): Prediction ID.
        tag (str): Predictor tag.
        type (PredictorType): Prediction type.
        created (str): Date created.
    """
    id: str
    tag: str
    type: PredictorType
    created: str
    FIELDS = f"""
    id
    tag
    type
    created
    ... on CloudPrediction {{
        results {{
            data
            type
            shape
        }}
        latency
        error
        logs
    }}
    """

    @classmethod
    def create (
        cls,
        tag: str,
        raw_outputs: bool=False,
        return_binary_path: bool=True,
        data_url_limit: int=None,
        access_key: str=None,
        **inputs: Dict[str, Union[ndarray, str, float, int, bool, List, Dict[str, Any], Path, Image.Image, Value]],
    ) -> Union[CloudPrediction, EdgePrediction]:
        """
        Create a prediction.

        Parameters:
            tag (str): Predictor tag.
            raw_outputs (bool): Skip converting output values into Pythonic types.
            return_binary_path (bool): Write binary values to file and return a `Path` instead of returning `BytesIO` instance.
            data_url_limit (int): Return a data URL if a given output value is smaller than this size in bytes. Only applies to `CLOUD` predictions.
            access_key (str): Function access key.
            inputs (dict): Input values. Only applies to `CLOUD` predictions.

        Returns:
            CloudPrediction | EdgePrediction: Created prediction.
        """
        # Collect inputs
        key = uuid4().hex
        inputs = { name: Value.from_value(value, name, key=key) for name, value in inputs.items() }
        inputs = [{ "name": name, **asdict(value) } for name, value in inputs.items()]
        # Query
        response = query(f"""
            mutation ($input: CreatePredictionInput!) {{
                createPrediction (input: $input) {{
                    {cls.FIELDS}
                }}
            }}""",
            { "input": { "tag": tag, "client": cls.__get_client(), "inputs": inputs, "dataUrlLimit": data_url_limit } },
            access_key=access_key
        )
        # Check
        prediction = response["createPrediction"]
        if not prediction:
            return None
        # Parse results
        if "results" in prediction and prediction["results"] is not None:
            prediction["results"] = [Value(**value) for value in prediction["results"]]
            if not raw_outputs:
                prediction["results"] = [value.to_value(return_binary_path=return_binary_path) for value in prediction["results"]]
        # Create
        prediction = CloudPrediction(**prediction) if prediction["type"] == PredictorType.Cloud else EdgePrediction(**prediction)
        # Return
        return prediction

    @classmethod
    def __get_client (cls) -> str:
        id = system()
        if id == "Darwin":
            return "macos"
        if id == "Linux":
            return "linux"
        if id == "Windows":
            return "windows"
        raise RuntimeError(f"Function cannot make predictions on the {id} platform")

@dataclass(frozen=True)
class CloudPrediction (Prediction):
    """
    Cloud prediction.

    Members:
        results (list): Prediction results.
        latency (float): Prediction latency in milliseconds.
        error (str): Prediction error. This is `null` if the prediction completed successfully.
        logs (str): Prediction logs.
    """
    results: List[Value] = None
    latency: float = None
    error: str = None
    logs: str = None

@dataclass(frozen=True)
class EdgePrediction (Prediction):
    """
    Edge prediction.
    """


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from os import environ

# Define API URL and access key
api_url = environ.get("FXN_API_URL", "https://api.fxn.ai/graph")
access_key: str = environ.get("FXN_ACCESS_KEY", None)

# Import everything
from .api import *
from .version import __version__


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import asdict
from io import BytesIO
from numpy import ndarray
from pathlib import Path, PurePath
from PIL import Image
from rich import print_json
from rich.progress import Progress, SpinnerColumn, TextColumn
from tempfile import mkstemp
from typer import Argument, Context, Option

from ..api import Prediction
from .auth import get_access_key

def predict (
    tag: str = Argument(..., help="Predictor tag."),
    raw_outputs: bool = Option(False, "--raw-outputs", help="Output raw Function values instead of converting into plain Python values."),
    context: Context = 0
):
    # Predict
    with Progress(
        SpinnerColumn(spinner_name="point"),
        TextColumn("[progress.description]{task.description}"),
        transient=True
    ) as progress:
        progress.add_task(description="Making prediction...", total=None)
        inputs = { context.args[i].replace("-", ""): _parse_value(context.args[i+1]) for i in range(0, len(context.args), 2) }
        prediction = Prediction.create(
            tag=tag,
            **inputs,
            raw_outputs=raw_outputs,
            return_binary_path=True,
            access_key=get_access_key()
        )
    # Parse results
    images = []
    if hasattr(prediction, "results") and prediction.results is not None:
        images = [value for value in prediction.results if isinstance(value, Image.Image)]
        results = [_serialize_value(value) for value in prediction.results]
        object.__setattr__(prediction, "results", results)
    # Print
    print_json(data=asdict(prediction, dict_factory=_prediction_dict_factory))
    # Show images
    for image in images:
        image.show()

def _parse_value (value: str):
    """
    Parse a value from a CLI argument.

    Parameters:
        value (str): CLI input argument.

    Returns:
        bool | int | float | str | Path: Parsed value.
    """
    # Boolean
    if value == "true":
        return True
    if value == "false":
        return False
    # Integer
    try:
        return int(value)
    except ValueError:
        pass
    # Float
    try:
        return float(value)
    except ValueError:
        pass
    # File
    if value.startswith("@"):
        return Path(value[1:])
    # String
    return value
    
def _serialize_value (value):
    # Convert ndarray to list
    if isinstance(value, ndarray):
        return value.tolist()
    # Write image
    if isinstance(value, Image.Image):
        _, path = mkstemp(suffix=".png" if value.mode == "RGBA" else ".jpg")
        value.save(path)
        return path
    # Serialize `BytesIO`
    if isinstance(value, BytesIO):
        return str(value)
    # Serialize `Path`
    if isinstance(value, PurePath):
        return str(value)
    # Return    
    return value

def _prediction_dict_factory (kv_pairs):
    # Check if value
    VALUE_KEYS = ["data", "type", "shape"]
    keys = [k for k, _ in kv_pairs]
    is_value = all(k in keys for k in VALUE_KEYS)
    kv_pairs = [(k, v) for k, v in kv_pairs if v is not None] if is_value else kv_pairs
    # Construct
    return dict(kv_pairs)


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from .api import query as queryfxn
from .dtype import Dtype
from .profile import Profile
from .storage import Storage, UploadType
from .value import Value

@dataclass(frozen=True)
class Predictor:
    """
    Predictor.

    Members:
        tag (str): Predictor tag.
        owner (Profile): Predictor owner.
        name (str): Predictor name.
        type (PredictorType): Predictor type.
        status (PredictorStatus): Predictor status.
        access (AccessMode): Predictor access.
        created (str): Date created.
        description (str): Predictor description.
        readme (str): Predictor readme. This supports Markdown.
        media (str): Predictor media URL.
        acceleration (Acceleration): Predictor acceleration. This only applies to cloud predictors.
        signature (Signature): Predictor signature. This is only populated once predictor has been successfully provisioned.
        license (str): Predictor license URL.
    """
    tag: str
    owner: Profile
    name: str
    type: PredictorType
    status: PredictorStatus
    access: AccessMode
    created: str
    description: Optional[str] = None
    readme: Optional[str] = None
    media: Optional[str] = None
    acceleration: Optional[Acceleration] = None
    signature: Optional[Signature] = None
    license: Optional[str] = None
    FIELDS = f"""
    tag
    owner {{
        {Profile.FIELDS}
    }}
    name
    type
    status
    access
    created
    description
    readme
    media
    acceleration
    signature {{
        inputs {{
            name
            type
            description
            range
            optional
            enumeration {{
                name
                value
            }}
            default_value: defaultValue {{
                data
                type
                shape
            }}
        }}
        outputs {{
            name
            type
            description
        }}
    }}
    license
    """

    def __post_init__ (self):
        owner = Profile(**self.owner, email=None) if isinstance(self.owner, dict) else self.owner
        signature = Signature(**self.signature) if isinstance(self.signature, dict) else self.signature
        object.__setattr__(self, "owner", owner)
        object.__setattr__(self, "signature", signature)

    @classmethod
    def retrieve (
        cls,
        tag: str,
        access_key: str=None
    ) -> Predictor:
        """
        Retrieve a predictor.

        Parameters:
            tag (str): Predictor tag.
            access_key (str): Function access key.

        Returns:
            Predictor: Predictor.
        """
        # Query
        response = queryfxn(f"""
            query ($input: PredictorInput!) {{
                predictor (input: $input) {{
                    {cls.FIELDS}
                }}
            }}
            """,
            { "input": { "tag": tag } },
            access_key=access_key
        )
        # Create predictor
        predictor = response["predictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor
    
    @classmethod
    def list (
        cls,
        owner: str=None,
        status: PredictorStatus=None,
        offset: int=None,
        count: int=None,
        access_key: str=None
    ) -> List[Predictor]:
        """
        List the current user's predictors.

        Parameters:
            owner (str): Predictor owner. This defaults to the current user.
            status (PredictorStatus): Predictor status. This defaults to `ACTIVE`.
            offset (int): Pagination offset.
            count (int): Pagination count.
            access_key (str): Function access key.

        Returns:
            list: User predictors.
        """
        # Query
        response = queryfxn(f"""
            query ($user: UserInput, $predictors: UserPredictorsInput) {{
                user (input: $user) {{
                    predictors (input: $predictors) {{
                        {cls.FIELDS}
                    }}
                }}
            }}
            """,
            {
                "user": { "username": owner } if owner else None,
                "predictors": { "status": status, "offset": offset, "count": count }
            },
            access_key=access_key
        )
        # Check
        user = response["user"]
        if not user:
            return None
        # Create predictors
        predictors = response["user"]["predictors"]
        predictors = [Predictor(**predictor) for predictor in predictors]
        # Return
        return predictors
    
    @classmethod
    def search (
        cls,
        query: str,
        offset: int=None,
        count: int=None,
        access_key: str=None
    ) -> List[Predictor]:
        """
        Search predictors.

        Parameters:
            q (str): Search query.
            offset (int): Pagination offset.
            count (int): Pagination count.
            access_key (str): Function access key.

        Returns:
            list: Relevant predictors.
        """
        # Query
        response = queryfxn(f"""
            query ($input: PredictorsInput!) {{
                predictors (input: $input) {{
                    {cls.FIELDS}
                }}
            }}
            """,
            { "input": { "query": query, "offset": offset, "count": count } },
            access_key=access_key
        )
        # Create predictors
        predictors = response["predictors"]
        predictors = [Predictor(**predictor) for predictor in predictors]
        # Return
        return predictors
    
    @classmethod
    def create (
        cls,
        tag: str,
        notebook: Union[str, Path],
        type: PredictorType=None,
        access: AccessMode=None,
        description: str=None,
        media: Union[str, Path]=None,
        acceleration: Acceleration=None,
        environment: Dict[str, str]=None,
        license: str=None,
        overwrite: bool=None,
        access_key: str=None
    ) -> Predictor:
        """
        Create a predictor.

        Parameters:
            tag (str): Predictor tag.
            notebook (str | Path): Predictor notebook path or URL.
            type (PredictorType): Predictor type. This defaults to `CLOUD`.
            access (AccessMode): Predictor access mode. This defaults to `PRIVATE`.
            description (str): Predictor description. This must be under 200 characters long.
            media (str | Path): Predictor media path or URL.
            acceleration (Acceleration): Predictor acceleration. This only applies for cloud predictors and defaults to `CPU`. 
            environment (dict): Predictor environment variables.
            license (str): Predictor license URL.
            overwrite (bool): Overwrite any existing predictor with the same tag. Existing predictor will be deleted.
            access_key (str): Function access key.

        Returns:
            Predictor: Created predictor.
        """
        # Prepare
        environment = [{ "name": name, "value": value } for name, value in environment.items()] if environment is not None else []
        notebook = Storage.upload(notebook, UploadType.Notebook) if isinstance(notebook, Path) else notebook
        media = Storage.upload(media, UploadType.Media) if isinstance(media, Path) else media
        # Query
        response = queryfxn(f"""
            mutation ($input: CreatePredictorInput!) {{
                createPredictor (input: $input) {{
                    {cls.FIELDS}
                }}
            }}
            """,
            {
                "input": {
                    "tag": tag,
                    "type": type,
                    "notebook": notebook,
                    "access": access,
                    "description": description,
                    "media": media,
                    "acceleration": acceleration,
                    "environment": environment,
                    "overwrite": overwrite,
                    "license": license
                }
            },
            access_key=access_key
        )
        # Create predictor
        predictor = response["createPredictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor
    
    @classmethod
    def delete (
        cls,
        tag: str,
        access_key: str=None
    ) -> bool:
        """
        Delete a predictor.

        Parameters:
            tag (str): Predictor tag.
            access_key (str): Function access key.

        Returns:
            bool: Whether the predictor was successfully deleted.
        """
        # Query
        response = queryfxn(f"""
            mutation ($input: DeletePredictorInput!) {{
                deletePredictor (input: $input)
            }}
            """,
            { "input": { "tag": tag } },
            access_key=access_key
        )
        # Return
        result = response["deletePredictor"]
        return result
    
    @classmethod
    def archive (
        cls,
        tag: str,
        access_key: str=None
    ) -> Predictor:
        """
        Archive an active predictor.

        Parameters:
            tag (str): Predictor tag.
            access_key (str): Function access key.

        Returns:
            Predictor: Archived predictor.
        """
        # Query
        response = queryfxn(f"""
            mutation ($input: ArchivePredictorInput!) {{
                archivePredictor (input: $input) {{
                    {cls.FIELDS}
                }}
            }}
            """,
            { "input": { "tag": tag } },
            access_key=access_key
        )
        # Create predictor
        predictor = response["archivePredictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor

@dataclass(frozen=True)
class Signature:
    """
    Predictor signature.

    Members:
        inputs (list): Input parameters.
        outputs (list): Output parameters.
    """
    inputs: List[Parameter]
    outputs: List[Parameter]

    def __post_init__ (self):
        inputs = [Parameter(**parameter) if isinstance(parameter, dict) else parameter for parameter in self.inputs]
        outputs = [Parameter(**parameter) if isinstance(parameter, dict) else parameter for parameter in self.outputs]
        object.__setattr__(self, "inputs", inputs)
        object.__setattr__(self, "outputs", outputs)

@dataclass(frozen=True)
class Parameter:
    """
    Predictor parameter.

    Members:
        name (str): Parameter name. This is only populated for input parameters.
        type (Dtype): Parameter type. This is `None` if the type is unknown or unsupported by Function.
        description (str): Parameter description.
        optional (bool): Parameter is optional.
        range (tuple): Parameter value range for numeric parameters.
        enumeration (list): Parameter value choices for enumeration parameters.
        default_value (str | float | int | bool): Parameter default value.
    """
    name: Optional[str] = None
    type: Optional[Dtype] = None
    description: Optional[str] = None
    optional: Optional[bool] = None
    range: Optional[Tuple[float, float]] = None
    enumeration: Optional[List[EnumerationMember]] = None
    default_value: Optional[Union[str, float, int, bool]] = None

    def __post_init__ (self):
        default_value = Value(**self.default_value).to_value() if isinstance(self.default_value, dict) and all(x in self.default_value for x in ["data", "type", "shape"]) else self.default_value
        enumeration = [EnumerationMember(**member) if isinstance(member, dict) else member for member in self.enumeration] if self.enumeration else self.enumeration
        object.__setattr__(self, "default_value", default_value)
        object.__setattr__(self, "enumeration", enumeration)

@dataclass(frozen=True)
class EnumerationMember:
    """
    Prediction parameter enumeration member.

    Members:
        name (str): Enumeration member name.
        value (str | int): Enumeration member value.
    """
    name: str
    value: Union[str, int]

class Acceleration (str, Enum):
    """
    Predictor acceleration.
    """
    CPU = "CPU"
    A40 = "A40"
    A100 = "A100"

class AccessMode (str, Enum):
    """
    Predictor access mode.
    """
    Public = "PUBLIC"
    Private = "PRIVATE"

class PredictorType (str, Enum):
    """
    Predictor type.
    """
    Cloud = "CLOUD"
    Edge = "EDGE"

class PredictorStatus (str, Enum):
    """
    Predictor status.
    """
    Provisioning = "PROVISIONING"
    Active = "ACTIVE"
    Invalid = "INVALID"
    Archived = "ARCHIVED"


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional

from .api import query

@dataclass(frozen=True)
class EnvironmentVariable:
    """
    Predictor environment variable.

    Members:
        name (str): Variable name.
        value (str): Variable value.
    """
    name: str
    value: Optional[str]=None
    FIELDS = f"""
    name
    """
    DEFAULT_VALUE = "xxxxxxxx"

    @classmethod
    def list (
        cls,
        organization: str=None,
        access_key: str=None
    ) -> List[EnvironmentVariable]:
        # Query
        response = query(f"""
            query ($input: UserInput) {{
                user (input: $input) {{
                    ... on User {{
                        environmentVariables {{
                            {cls.FIELDS}
                        }}
                    }}
                    ... on Organization {{
                        environmentVariables {{
                            {cls.FIELDS}
                        }}
                    }}
                }}
            }}
            """,
            { "input": { "username": organization } if organization is not None else None },
            access_key=access_key
        )
        # Create envs
        assert response["user"] is not None, "Failed to list environment variables because user could not be found. Check that you are authenticated."
        environments = response["user"]["environmentVariables"]
        environments = [EnvironmentVariable(**env, value=cls.DEFAULT_VALUE) for env in environments]
        # Return
        return environments

    @classmethod
    def create (
        cls,
        name: str,
        value: str,
        organization: str=None,
        access_key: str=None
    ) -> EnvironmentVariable:
        """
        Create an environment variable.

        This environment variable will apply to all predictors you create.

        Parameters:
            name (str): Variable name.
            value (str): Variable value.
            organization (str): Organization username. Use this for organization environment variables.
            access_key (str): Function access key.

        Returns:
            EnvironmentVariable: Created environment variable.
        """
        # Query
        response = query(f"""
            mutation ($input: CreateEnvironmentVariableInput!) {{
                environment: createEnvironmentVariable (input: $input) {{
                    {cls.FIELDS}
                }}
            }}
            """,
            { "input": { "name": name, "value": value, "organization": organization } },
            access_key=access_key
        )
        # Create env
        environment = response["environment"]
        environment = EnvironmentVariable(**environment, value=cls.DEFAULT_VALUE)
        # Return
        return environment

    @classmethod
    def delete (
        cls,
        name: str,
        organization: str=None,
        access_key: str=None
    ) -> bool:
        """
        Delete an environment variable.

        Parameters:
            name (str): Variable name.
            organization (str): Organization username. Use this for organization environment variables.
            access_key (str): Function access key.
        
        Returns:
            bool: Whether the environment variable was successfully deleted.
        """
        # Query
        response = query(f"""
            mutation ($input: DeleteEnvironmentVariableInput!) {{
                result: deleteEnvironmentVariable (input: $input)
            }}
            """,
            { "input": { "name": name, "organization": organization } },
            access_key=access_key
        )
        # Return
        result = response["result"]
        return result


#
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import dataclass

@dataclass(frozen=True)
class Tag:
    """
    Predictor tag.

    Members:
        username (str): Predictor owner username.
        name (str): Predictor name.
    """
    username: str
    name: str

def parse_tag (tag: str) -> Tag:
    """
    Parse a predictor tag.

    Parameters:
        tag (str): Tag string.

    Returns:
        Tag: Parsed tag.
    """
    username, name = tag.lower()[1:].split("/")
    result = Tag(username=username, name=name)
    return result

def serialize_tag (tag: Tag) -> str:
    """
    Serialize a predictor tag.

    Parameters:
        tag (Tag): Tag.

    Returns:
        str: Serialized tag.
    """
    username, name = tag
    result = f"@{username}/{name}"
    return result


#
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from dataclasses import asdict, dataclass, is_dataclass
from filetype import guess_mime
from io import BytesIO
from json import loads, dumps
from numpy import array, float32, frombuffer, int32, ndarray
from pathlib import Path
from PIL import Image
from requests import get
from tempfile import NamedTemporaryFile
from typing import Dict, List, Optional, Union
from urllib.request import urlopen

from .dtype import Dtype
from .storage import Storage, UploadType

@dataclass(frozen=True)
class Value:
    """
    Prediction value.

    Members:
        data (str): Value URL. This can be a web URL or a data URL.
        type (Dtype): Value type.
        shape (list): Value shape. This is `None` if shape information is not available or applicable.
    """
    data: str
    type: Dtype
    shape: Optional[List[int]] = None

    def to_value (
        self,
        return_binary_path: bool=True
    ) -> Union[str, float, int, bool, ndarray, list, dict, Image.Image, BytesIO, Path]:
        """
        Convert a Function value to a plain value.

        Parameters:
            return_binary_path (str): Write binary values to file and return a `Path` instead of returning `BytesIO` instance.

        Returns:
            str | float | int | bool | ndarray | list | dict | Image.Image | BytesIO | Path: Value.
        """
        # Null
        if self.type == Dtype.null:
            return None
        # Download
        buffer = Value.__download_value_data(self.data)
        # Array
        if self.type in [
            Dtype.int8, Dtype.int16, Dtype.int32, Dtype.int64,
            Dtype.uint8, Dtype.uint16, Dtype.uint32, Dtype.uint64,
            Dtype.float16, Dtype.float32, Dtype.float64, Dtype.bool
        ]:
            assert self.shape is not None, "Array value must have a shape specified"
            array = frombuffer(buffer.getbuffer(), dtype=self.type).reshape(self.shape)
            return array if len(self.shape) > 0 else array.item()
        # String
        if self.type == Dtype.string:
            return buffer.getvalue().decode("utf-8")
        # List
        if self.type == Dtype.list:
            return loads(buffer.getvalue().decode("utf-8"))
        # Dict
        if self.type == Dtype.dict:
            return loads(buffer.getvalue().decode("utf-8"))
        # Image
        if self.type == Dtype.image:
            return Image.open(buffer)
        # Binary
        if return_binary_path:
            with NamedTemporaryFile(mode="wb", delete=False) as f:
                f.write(buffer.getbuffer())
            return Path(f.name)
        # Return
        return buffer

    @classmethod
    def from_value (
        cls,
        value: Union[str, float, int, bool, ndarray, List, Dict[str, any], Path, Image.Image],
        name: str,
        min_upload_size: int=4096,
        key: str=None
    ) -> Value:
        """
        Create a Function value from a plain value.

        Parameters:
            value (str | float | int | bool | ndarray | list | dict | dataclass | Path | PIL.Image): Input value.
            name (str): Value name.
            min_upload_size (int): Values larger than this size in bytes will be uploaded.

        Returns:
            Value: Function value.
        """
        # None
        if value is None:
            return Value(None, type=Dtype.null)
        # Value
        if isinstance(value, Value):
            return value
        # Array
        if isinstance(value, ndarray):
            buffer = BytesIO(value.tobytes())
            data = Storage.upload(buffer, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data, type=value.dtype.name, shape=list(value.shape))
        # String
        if isinstance(value, str):
            buffer = BytesIO(value.encode("utf-8"))
            data = Storage.upload(buffer, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data, type=Dtype.string)
        # Float
        if isinstance(value, float):
            value = array(value, dtype=float32)
            return Value.from_value(value, name, min_upload_size=min_upload_size, key=key)
        # Boolean
        if isinstance(value, bool):
            value = array(value, dtype=bool)
            return Value.from_value(value, name, min_upload_size=min_upload_size, key=key)
        # Integer
        if isinstance(value, int):
            value = array(value, dtype=int32)
            return Value.from_value(value, name, min_upload_size=min_upload_size, key=key)
        # List
        if isinstance(value, list):
            value = dumps(value)
            buffer = BytesIO(value.encode("utf-8"))
            data = Storage.upload(buffer, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data, type=Dtype.list)
        # Dict
        if isinstance(value, dict):
            value = dumps(value)
            buffer = BytesIO(value.encode("utf-8"))
            data = Storage.upload(buffer, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data, type=Dtype.dict)
        # Dataclass # https://docs.python.org/3/library/dataclasses.html#dataclasses.is_dataclass
        if is_dataclass(value) and not isinstance(value, type):
            value = asdict(value)
            return Value.from_value(value, name=name, min_upload_size=min_upload_size, key=key)
        # Image
        if isinstance(value, Image.Image):
            buffer = BytesIO()
            format = "PNG" if value.mode == "RGBA" else "JPEG"
            value.save(buffer, format=format)
            data = Storage.upload(buffer, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data, type=Dtype.image)
        # Path
        if isinstance(value, Path):
            assert value.exists(), "Value does not exist at the given path"
            assert value.is_file(), "Value path must point to a file, not a directory"
            value = value.expanduser().resolve()
            data = Storage.upload(value, UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            type = Value.__get_file_dtype(value)
            return Value(data, type=type)
        # Unsupported
        raise RuntimeError(f"Cannot create Function value '{name}' for value {value} of type {type(value)}")

    @classmethod
    def __get_file_dtype (cls, path: Path) -> Dtype:
        mime = guess_mime(str(path))
        if not mime:
            return Dtype.binary
        if mime.startswith("image"):
            return Dtype.image
        if mime.startswith("video"):
            return Dtype.video
        if mime.startswith("audio"):
            return Dtype.audio
        if path.suffix in [".obj", ".gltf", ".glb", ".fbx", ".usd", ".usdz", ".blend"]:
            return Dtype._3d
        return Dtype.binary

    @classmethod
    def __download_value_data (cls, url: str) -> BytesIO:
        # Check if data URL
        if url.startswith("data:"):
            with urlopen(url) as response:
                return BytesIO(response.read())
        # Download
        response = get(url)
        result = BytesIO(response.content)
        return result


# 
#   Function
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from .dtype import Dtype
from .environment import EnvironmentVariable
from .prediction import CloudPrediction, EdgePrediction, Prediction
from .predictor import Acceleration, AccessMode, EnumerationMember, Parameter, Predictor, PredictorStatus, PredictorType, Signature
from .profile import Profile
from .storage import Storage, UploadType
from .user import User
from .value import Value